@@ -23,6 +23,22 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+# boilerplate to allow running as script directly
+if __name__ == "__main__" and __package__ is None:
+    import sys
+    import os
+    # The following assumes the script is in the top level of the package
+    # directory.  We use dirname() to help get the parent directory to add to
+    # sys.path, so that we can import the current package.  This is necessary
+    # since when invoked directly, the 'current' package is not automatically
+    # imported.
+    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
+    sys.path.insert(0, parent_dir)
+    import flootty
+    assert flootty
+    __package__ = str("flootty")
+    del sys, os
+
 import atexit
 import fcntl
 import json
@@ -41,6 +57,16 @@ import signal
 import re
 import collections
 
+PY2 = sys.version_info < (3, 0)
+
+
+try:
+    import __builtin__
+    input = getattr(__builtin__, 'raw_input')
+except (ImportError, AttributeError):
+    pass
+
+
 try:
     from urllib.parse import urlparse
     assert urlparse
@@ -48,11 +74,12 @@ except ImportError:
     from urlparse import urlparse
 
 try:
-    from . import api, cert, utils
-    assert api and cert and utils
+    from . import api, cert, shared as G, utils
+    assert api and cert and G and utils
 except (ImportError, ValueError):
     import api
     import cert
+    import shared as G
     import utils
 
 
@@ -86,9 +113,10 @@ def read_floorc():
 
 
 def write(fd, b):
+    if (not PY2) and isinstance(b, str):
+            b = b.encode('utf-8')
     while len(b):
         try:
-            # TODO: fix this for python3
             n = os.write(fd, b)
             b = b[n:]
         except (IOError, OSError):
@@ -96,7 +124,7 @@ def write(fd, b):
 
 
 def read(fd):
-    buf = ''
+    buf = b''
     while True:
         try:
             d = os.read(fd, FD_READ_BYTES)
@@ -207,6 +235,9 @@ def main():
 
     options, args = parser.parse_args()
 
+    G.USERNAME = options.username
+    G.SECRET = options.secret
+
     default_term_name = ""
     if options.create:
         default_term_name = "_"
@@ -234,6 +265,17 @@ def main():
         if not options.host:
             options.host = floo.get('host')
 
+    if not options.room or not options.owner:
+        try:
+            now_editing = api.get_now_editing_workspaces()
+            now_editing = json.loads(now_editing.read().decode('utf-8'))
+            if len(now_editing) == 1:
+                options.room = now_editing[0]['name']
+                options.owner = now_editing[0]['owner']
+            # TODO: list possible workspaces to join if > 1 is active
+        except Exception as e:
+            raise e
+
     if options.list:
         if len(term_name) != 0:
             die("I don't understand why you gave me a positional argument.")
@@ -287,7 +329,7 @@ class Flootty(object):
         self.reconnect_timeout = None
 
         self.buf_out = collections.deque()
-        self.buf_in = ''
+        self.buf_in = b''
 
         self.host = options.host
         self.port = int(options.port)
@@ -316,6 +358,15 @@ class Flootty(object):
         if fd.errer:
             self.errers.add(fileno)
 
+    def remove_fd(self, fileno):
+        self.readers.discard(fileno)
+        self.writers.discard(fileno)
+        self.errers.discard(fileno)
+        try:
+            del self.fds[fileno]
+        except KeyError:
+            pass
+
     def transport(self, name, data):
         data['name'] = name
         self.buf_out.append(data)
@@ -348,14 +399,18 @@ class Flootty(object):
             for position, fds in enumerate([_except, _in, _out]):
                 attr = attrs[position]
                 for fd in fds:
-                    handler = self.fds[fd][attr]
+                    # the handler can remove itself from self.fds...
+                    handler = self.fds.get(fd)
+                    if handler is None:
+                        continue
+                    handler = handler[attr]
                     if handler:
                         handler(fd)
                     else:
                         raise Exception('no handler for fd: %s %s' % (fd, attr))
 
     def cloud_read(self, fd):
-        buf = ''
+        buf = b''
         try:
             while True:
                 d = self.sock.recv(FD_READ_BYTES)
@@ -378,10 +433,11 @@ class Flootty(object):
         try:
             while True:
                 item = self.buf_out.popleft()
-                if self.authed:
-                    self.sock.sendall((json.dumps(item) + '\n').encode('utf-8'))
-                elif item['name'] == 'auth':
-                    self.sock.sendall((json.dumps(item) + '\n').encode('utf-8'))
+                data = json.dumps(item) + '\n'
+                if self.authed or item['name'] == 'auth':
+                    if not PY2:
+                        data = data.encode('utf-8')
+                    self.sock.sendall(data)
                 else:
                     new_buf_out.append(item)
         except socket.error:
@@ -398,10 +454,10 @@ class Flootty(object):
     def handle(self, req):
         self.buf_in += req
         while True:
-            before, sep, after = self.buf_in.partition('\n')
+            before, sep, after = self.buf_in.partition(b'\n')
             if not sep:
                 break
-            data = json.loads(before, encoding='utf-8')
+            data = json.loads(before.decode('utf-8'), encoding='utf-8')
             self.handle_event(data)
             self.buf_in = after
 
@@ -437,7 +493,7 @@ class Flootty(object):
         elif not self.term_name:
             if len(ri['terms']) == 0:
                 out('There is no active terminal in this workspace. Do you want to share your terminal? (y/n)')
-                choice = raw_input().lower()
+                choice = input().lower()
                 self.term_name = "_"
                 if choice == 'y':
                     self.options.create = True
@@ -446,7 +502,7 @@ class Flootty(object):
                 else:
                     die('If you ever change your mind, you can share your terminal using the --create [super_awesome_name] flag.')
             elif len(ri['terms']) == 1:
-                term_id, term = ri['terms'].items()[0]
+                term_id, term = list(ri['terms'].items())[0]
                 self.term_id = int(term_id)
                 self.term_name = term['term_name']
             else:
@@ -523,9 +579,7 @@ class Flootty(object):
         self.buf_out = new_buf_out
 
         if self.sock:
-            self.readers.remove(self.sock.fileno())
-            self.writers.remove(self.sock.fileno())
-            self.errers.remove(self.sock.fileno())
+            self.remove_fd(self.sock.fileno())
             try:
                 self.sock.shutdown(2)
             except Exception:
@@ -571,7 +625,7 @@ class Flootty(object):
                 self.sock.do_handshake()
         except socket.error as e:
             out('Error connecting: %s.' % e)
-            self.reconnect()
+            return self.reconnect()
         self.sock.setblocking(0)
         out('Connected!')
         self.send_auth()
@@ -607,6 +661,8 @@ class Flootty(object):
         def ship_stdin(fd):
             data = read(fd)
             if data:
+                if not PY2:
+                    data = data.decode('utf-8')
                 self.transport("term_stdin", {'data': data, 'id': self.term_id})
 
         if 'term_stdin' in self.ri['perms']:
@@ -616,7 +672,7 @@ class Flootty(object):
             out('You do not have permission to write to this terminal.')
 
         def stdout_write(buf):
-            write(stdout, buf.encode('utf-8'))
+            write(stdout, buf)
 
         self.handle_stdio = stdout_write
         self._set_pty_size(self.ri['terms'][str(self.term_id)]['size'])
@@ -626,12 +682,13 @@ class Flootty(object):
         Create a spawned process.
         Based on the code for pty.spawn().
         '''
-        out('Successfully joined %s' % (self.room_url()))
 
         if self.master_fd:
             # reconnected. don't spawn a new shell
+            out('Reconnected to %s' % (self.room_url()))
             return
         shell = os.environ['SHELL']
+        out('Successfully joined %s' % (self.room_url()))
 
         self.child_pid, self.master_fd = pty.fork()
         if self.child_pid == pty.CHILD:
@@ -645,27 +702,27 @@ class Flootty(object):
         def slave_death(fd):
             die('Exiting flootty because child exited.')
 
-        self.extra_data = ''
+        self.extra_data = b''
 
         def stdout_write(fd):
             '''
             Called when there is data to be sent from the child process back to the user.
             '''
             data = self.extra_data + os.read(fd, FD_READ_BYTES)
-            self.extra_data = ""
+            self.extra_data = b''
             if data:
                 while True:
                     try:
                         data.decode('utf-8')
                     except UnicodeDecodeError:
-                        self.extra_data = data[-1] + self.extra_data
+                        self.extra_data = data[-1:] + self.extra_data
                         data = data[:-1]
                     else:
                         break
                     if len(self.extra_data) > 100:
                         die('not a valid utf-8 string: %s' % self.extra_data)
                 if data:
-                    self.transport("term_stdout", {'data': data, 'id': self.term_id})
+                    self.transport("term_stdout", {'data': data.decode('utf-8'), 'id': self.term_id})
                     write(pty.STDOUT_FILENO, data)
 
         self.add_fd(self.master_fd, reader=stdout_write, errer=slave_death, name='create_term_stdout_write')
@@ -674,7 +731,7 @@ class Flootty(object):
             data = os.read(fd, FD_READ_BYTES)
             if data:
                 write(self.master_fd, data)
-                self.transport("term_stdin", {'data': data, 'id': self.term_id})
+                self.transport("term_stdin", {'data': data.decode('utf-8'), 'id': self.term_id})
 
         self.add_fd(pty.STDIN_FILENO, reader=stdin_write, name='create_term_stdin_write')
 
